/*
** Copyright (C) 2005-2008 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
** rwmatch
**
** rwmatch.c is a utility which takes two files (now called in and
** out) and produces a new file where the contents are linked together
** in the form (saddress, values) on one side and (daddress, values)
** on the other side.  apart from the saddress/daddress relationship,
** any other mating rules are specified by the command line.
**
** This is necessary because there are several operations where mating
** is not direct reversal; the particular case which is driving this
** applicatin is traceroute where the only rational is close
** relationship in time between s and d address, and one of them is a
** UDP signal with an ICMP response
**
** command arguments:
**
** --relate:    takes fields as pairs for relationship.  The first
**              value is the field in file 1, the second value is the
**              field in file ** 2.
**
**              for example, the default could be --relate=1,2
**
**              multiple --relate predicates are accepted.
**              Traditional tcp/ip matching would be:
**              --relate=1,2 --relate=2,1 --relate=3,4 --relate=4,3
**
** --delta:     the number of seconds difference we'll allow between
**              two records before considering them similar.  Defaults
**              to 0.
**
** --unmatched: Include unmatched queries in match file.  This is a
**              less than general solution, but allows scans with scan
**              responses to be captured in a single file.
**
** --unmated:   output path for unmatched records; defaults to
**              /dev/null.  A record must have a response before it is
**              ejected.  So we're always sending out at least two
**              records.
**
** Apart from that; rwmate will annotate the nhip field with an index
** indicating the 'mating flow' that a record belongs to.
**
*/

#include "silk.h"

RCSIDENT("$SiLK: rwmatch.c 10955 2008-03-19 16:54:34Z mthomas $");

#include "utils.h"
#include "rwpack.h"
#include "sksite.h"


/* TYPEDEFS AND DEFINES */

/* Where to send --help output */
#define USAGE_FH stdout

#define RWMATE_REL_MAX 16
#define RWMATE_REL_MAX_STR "16"

#define RWMATE_QUERY 0
#define RWMATE_RESPONSE 1


typedef enum {
    RWMATE_FID_SIP = 1, RWMATE_FID_DIP, RWMATE_FID_SPORT, RWMATE_FID_DPORT,
    RWMATE_FID_PROTO, RWMATE_FID_PKTS, RWMATE_FID_BYTES, RWMATE_FID_FLAGS
} fields_enum_t;


/* LOCAL VARIABLES */

/* input and output streams */
static rwIOStruct_t *query_stream = NULL;
static rwIOStruct_t *response_stream = NULL;
static rwIOStruct_t *matched_stream = NULL;

/* the pairs of fields to match on */
static uint8_t count_pairs;
static uint8_t match_pairs[RWMATE_REL_MAX][2];

/* time difference between query and response in milliseconds */
static sktime_t delta_msec = 0;

/* whether to allow the "response" to precede the "query"; for
 * handling traffic where either side may initiate the connection. */
static int symmetric_delta = 0;

/* whether to write unmatched queries/responses in match file */
static int write_unmatched_query = 0;
static int write_unmatched_response = 0;


/* OPTIONS SETUP */

typedef enum {
    OPT_RELATE, OPT_TIME_DELTA, OPT_UNMATCHED, OPT_SYMMETRIC_DELTA
} appOptionsEnum;

static struct option appOptions[] = {
    {"relate",          REQUIRED_ARG, 0, OPT_RELATE},
    {"time-delta",      REQUIRED_ARG, 0, OPT_TIME_DELTA},
    {"symmetric-delta", NO_ARG,       0, OPT_SYMMETRIC_DELTA},
    {"unmatched",       REQUIRED_ARG, 0, OPT_UNMATCHED},
    {0,0,0,0}           /* sentinel entry */
};

static const char *appHelp[] = {
    ("A pair of fields to relate when matching data, given in\n"
     "\tthe form '<query field id>,<response field id>'.  The switch may\n"
     "\tbe repeated up to " RWMATE_REL_MAX_STR " times.  Field IDs:\n"
     "\t    1 source IP          5 protocol    \n"
     "\t    2 destination IP     6 packets     \n"
     "\t    3 source port        7 bytes       \n"
     "\t    4 destination port   8 TCP flags   "),
    ("Maximum time difference (in seconds) permitted between\n"
     "\ttwo records for relation. Def. 0"),
    ("Also match responses the precede queries by up to\n"
     "\ttime-delta seconds. Def. No"),
    ("Include unmatched records from QUERY_FILE and/or\n"
     "\tRESPONSE_FILE in MATCH_FILE.  Parameter is one of [QqRrBb], where:\n"
     "\tQ / q - query file; R / r - response file, B /b - both"),
    (char *)NULL
};


/* LOCAL FUNCTION PROTOTYPES */

static void appUsageLong(void);
static void appTeardown(void);
static void appSetup(int argc, char **argv);
static int  appOptionsHandler(clientData cData, int opt_index, char *opt_arg);
static uint32_t getField(fields_enum_t field_id, const rwRec *current_rec);
static int  compareRecords(const rwRec *query_rec, const rwRec *response_rec);


/* FUNCTION DEFINITIONS */

/*
 *  appUsageLong();
 *
 *    Print complete usage information to USAGE_FH.  Pass this
 *    function to skOptionsSetUsageCallback(); skOptionsParse() will
 *    call this funciton and then exit the program when the --help
 *    option is given.
 */
static void appUsageLong(void)
{
#define USAGE_MSG                                                       \
    ("--relate=FIELD_PAIR QUERY_FILE RESPONSE_FILE MATCHED_PATH\n"      \
     "\tGroups records as queries and responses\n")

    FILE *fh = USAGE_FH;

    skAppStandardUsage(fh, USAGE_MSG, appOptions, appHelp);
    sksiteOptionsUsage(fh);
}


/*
 *  appSetup(argc, argv);
 *
 *    Perform all the setup for this application include setting up
 *    required modules, parsing options, etc.  This function should be
 *    passed the same arguments that were passed into main().
 *
 *    Returns to the caller if all setup succeeds.  If anything fails,
 *    this function will cause the application to exit with a FAILURE
 *    exit status.
 */
static void appSetup(int argc, char **argv)
{
    int arg_index;
    int rv;

    /* verify same number of options and help strings */
    assert((sizeof(appHelp)/sizeof(char *)) ==
           (sizeof(appOptions)/sizeof(struct option)));

    /* register the application */
    skAppRegister(argv[0]);
    skOptionsSetUsageCallback(&appUsageLong);

    /* initialize globals */
    count_pairs = 0;

    /* register the options */
    if (skOptionsRegister(appOptions, &appOptionsHandler, NULL)
        || sksiteOptionsRegister(SK_SITE_FLAG_CONFIG_FILE))
    {
        skAppPrintErr("Unable to register options");
        exit(EXIT_FAILURE);
    }

    /* parse options */
    arg_index = skOptionsParse(argc, argv);
    if (arg_index < 0) {
        skAppUsage();             /* never returns */
    }

    /* try to load site config file; if it fails, we will not be able
     * to resolve flowtype and sensor from input file names */
    sksiteConfigure(0);

    /* verify that related pairs were given */
    if (count_pairs == 0) {
        skAppPrintErr("At least one --%s pair must be given",
                      appOptions[OPT_RELATE].name);
        skAppUsage();             /* never returns */
    }

    /* get the file arguments */
    if (arg_index == argc) {
        skAppPrintErr("Missing QUERY_FILE argument");
        skAppUsage();             /* never returns */
    }
    rv = rwioCreate(&query_stream, argv[arg_index], SK_IO_READ);
    if (rv) {
        rwioPrintLastErr(query_stream, rv, &skAppPrintErr);
        exit(EXIT_FAILURE);
    }
    arg_index++;

    if (arg_index == argc) {
        skAppPrintErr("Missing RESPONSE_FILE argument");
        skAppUsage();             /* never returns */
    }
    rv = rwioCreate(&response_stream, argv[arg_index], SK_IO_READ);
    if (rv) {
        rwioPrintLastErr(response_stream, rv, &skAppPrintErr);
        exit(EXIT_FAILURE);
    }
    arg_index++;

    if (arg_index == argc) {
        skAppPrintErr("Missing MATCHED_FILE argument");
        skAppUsage();             /* never returns */
    }
    rv = rwioCreate(&matched_stream, argv[arg_index], SK_IO_WRITE);
    if (rv) {
        rwioPrintLastErr(matched_stream, rv, &skAppPrintErr);
        exit(EXIT_FAILURE);
    }
    arg_index++;

    /* check for extra options */
    if (arg_index != argc) {
        skAppPrintErr("Too many arguments or unrecognized switch '%s'",
                      argv[arg_index]);
        skAppUsage(); /* never returns */
    }

    /* Now, open the files */
    if ((rv = skStreamSetIPv6Policy(query_stream, SK_IPV6POLICY_ASV4))
        || (rv = rwioOpen(query_stream)))
    {
        rwioPrintLastErr(query_stream, rv, &skAppPrintErr);
        skAppPrintErr("Cannot open QUERY_FILE. Exiting.");
        exit(EXIT_FAILURE);
    }

    if ((rv = skStreamSetIPv6Policy(response_stream, SK_IPV6POLICY_ASV4))
        || (rv = rwioOpen(response_stream)))
    {
        rwioPrintLastErr(response_stream, rv, &skAppPrintErr);
        skAppPrintErr("Cannot open RESPONSE_FILE. Exiting.");
        exit(EXIT_FAILURE);
    }

    if ((rv = rwioOpen(matched_stream))
        || (rv = rwioWriteHeader(matched_stream)))
    {
        rwioPrintLastErr(matched_stream, rv, &skAppPrintErr);
        skAppPrintErr("Cannot open MATCHED_FILE. Exiting.");
        exit(EXIT_FAILURE);
    }

    if (atexit(appTeardown) < 0) {
        skAppPrintErr("Unable to register appTeardown() with atexit()");
        appTeardown();
        exit(EXIT_FAILURE);
    }

    return;                     /* OK */
}


/*
 *  appTeardown()
 *
 *    Teardown all modules, close all files, and tidy up all
 *    application state.
 *
 *    This function is idempotent.
 */
static void appTeardown(void)
{
    static int teardownFlag = 0;
    int rv;

    if (teardownFlag) {
        return;
    }
    teardownFlag = 1;

    if (query_stream != NULL) {
        rv = rwCloseFile(query_stream);
        if (rv) {
            rwioPrintLastErr(query_stream, rv, &skAppPrintErr);
        }
    }
    if (response_stream != NULL) {
        rv = rwCloseFile(response_stream);
        if (rv) {
            rwioPrintLastErr(response_stream, rv, &skAppPrintErr);
        }
    }
    if (matched_stream != NULL) {
        rv = rwCloseFile(matched_stream);
        if (rv) {
            rwioPrintLastErr(matched_stream, rv, &skAppPrintErr);
        }
    }

    skAppUnregister();
}


/*
 *  status = appOptionsHandler(cData, opt_index, opt_arg);
 *
 *    This function is passed to skOptionsRegister(); it will be called
 *    by skOptionsParse() for each user-specified switch that the
 *    application has registered; it should handle the switch as
 *    required---typically by setting global variables---and return 1
 *    if the switch processing failed or 0 if it succeeded.  Returning
 *    a non-zero from from the handler causes skOptionsParse() to return
 *    a negative value.
 *
 *    The clientData in 'cData' is typically ignored; 'opt_index' is
 *    the index number that was specified as the last value for each
 *    struct option in appOptions[]; 'opt_arg' is the user's argument
 *    to the switch for options that have a REQUIRED_ARG or an
 *    OPTIONAL_ARG.
 */
static int appOptionsHandler(
    clientData  UNUSED(cData),
    int         opt_index,
    char       *opt_arg)
{
    uint32_t *parsed_list;
    uint32_t count;
    int rv;

    switch ((appOptionsEnum)opt_index) {
      case OPT_RELATE:
        if (count_pairs == RWMATE_REL_MAX) {
            skAppPrintErr(("Total number of relations specified"
                           " exceeds maximum (%d)"),
                          RWMATE_REL_MAX);
            return 1;
        }
        rv = skStringParseNumberList(&parsed_list, &count, opt_arg,
                                     RWMATE_FID_SIP, RWMATE_FID_FLAGS, 2);
        if (rv) {
            skAppPrintErr("Invalid %s '%s': %s",
                          appOptions[opt_index].name, opt_arg,
                          skStringParseStrerror(rv));
            return 1;
        }
        assert(count > 0 && count <= 2);
        if (count != 2) {
            skAppPrintErr("Invalid %s '%s': Each relation requires two IDs",
                          appOptions[opt_index].name, opt_arg);
            free(parsed_list);
            return 1;
        }
        match_pairs[count_pairs][RWMATE_QUERY] = parsed_list[0];
        match_pairs[count_pairs][RWMATE_RESPONSE] = parsed_list[1];
        count_pairs++;
        free(parsed_list);
        break;

      case OPT_TIME_DELTA:
        rv = skStringParseUint32(&count, opt_arg, 0, 0);
        if (rv) {
            skAppPrintErr("Invalid %s '%s': %s",
                          appOptions[opt_index].name, opt_arg,
                          skStringParseStrerror(rv));
            return 1;
        }
        delta_msec = (sktime_t)count * (sktime_t)1000;
        break;

      case OPT_SYMMETRIC_DELTA:
        symmetric_delta = 1;
        break;

      case OPT_UNMATCHED:
          if (strlen(opt_arg) != 1) {
              skAppPrintErr(("Invalid %s '%s': "
                             "argument must be one of \"q,r,b\""),
                            appOptions[opt_index].name, opt_arg);
              return 1;
          } else {
              switch (opt_arg[0]) {
                case 'Q':
                case 'q':
                  write_unmatched_query = 1;
                  break;
                case 'R':
                case 'r':
                  write_unmatched_response = 1;
                  break;
                case 'B':
                case 'b':
                  write_unmatched_query = 1;
                  write_unmatched_response = 1;
                  break;
                default:
                  skAppPrintErr(("Invalid %s '%s': "
                                 "argument must be one of \"qrb\""),
                                appOptions[opt_index].name, opt_arg);
                  return 1;
              }
          }
          break;
    }

    return 0; /* OK */
}


/*
 * uint32_t getField(fields_enum_t current_field, rwRec *current_rec)
 *
 * Given a record and a field ID returns the value of the field for
 * that record.
 */
static uint32_t getField(fields_enum_t field_id, const rwRec *current_rec)
{
    uint32_t result = 0;

    switch (field_id) {
      case RWMATE_FID_SIP:
        result = rwRecGetSIPv4(current_rec);
        break;
      case RWMATE_FID_DIP:
        result = rwRecGetDIPv4(current_rec);
        break;
      case RWMATE_FID_SPORT:
        result = rwRecGetSPort(current_rec);
        break;
      case RWMATE_FID_DPORT:
        result = rwRecGetDPort(current_rec);
        break;
      case RWMATE_FID_PROTO:
        result = rwRecGetProto(current_rec);
        break;
      case RWMATE_FID_PKTS:
        result = rwRecGetPkts(current_rec);
        break;
      case RWMATE_FID_BYTES:
        result = rwRecGetBytes(current_rec);
        break;
      case RWMATE_FID_FLAGS:
        result = rwRecGetFlags(current_rec);
        break;
    }
    return result;
}


/*
 * int compareRecords(rwRec *propose, rwRec *response)
 *
 * this uses the same hash information from rwm_hash to compare an
 * query and response record in order to determine whether or not.
 * Note the global variable dependence.
 *
 * Returns 1 if the query > outing, -1 if response > query, 0 if
 * equal
 */
static int compareRecords(
    const rwRec    *query_rec,
    const rwRec    *response_rec)
{
    uint8_t i;
    uint32_t qval;
    uint32_t rval;

    for (i = 0 ; i < count_pairs; i++) {
        qval = getField(match_pairs[i][RWMATE_QUERY], query_rec);
        rval = getField(match_pairs[i][RWMATE_RESPONSE], response_rec);
        if (qval > rval) {
            return 1;
        } else if (qval < rval) {
            return -1;
        }
    }

    /*
     * We successfully matched all match fields that we care about.  Now we
     * compare timing.
     *
     * The response should be within <delta> seconds of the end of the
     * message.
     *
     * If we are making the match window symmetric about the query
     * time, adjust the match start.
     */
    if ((rwRecGetStartTime(query_rec) - (symmetric_delta ? delta_msec : 0))
        > rwRecGetStartTime(response_rec))
    {
        return 1;
    }

    /*
     * We know response.etime > outging.stime, and we know by now that
     * query.stime >= response.stime; so we just have to compare etime.
     */
    if ((rwRecGetEndTime(query_rec) + delta_msec)
        < rwRecGetEndTime(response_rec))
    {
        return -1;
    }

    /*
     * we've passed all tests, so the two must be related.
     */
    return 0;
}


int main(int argc, char **argv)
{
    rwRec query_rec;
    rwRec response_rec;
    uint32_t match_id = 0;
    int printed_query = 0;
    int have_query = 1;
    int have_response = 1;
    int rv;

    appSetup(argc, argv); /* never returns on error */

    /*
     * The revised version of this application requires
     * sorted data that matches the mating.  So this means
     * that we're going to have records sorted slightly
     * differently (if we match 2/1, we sort --field=2,9 and
     * --field=1,9) on the two applications.
     *
     * The loop always begins with us
     * with a reference input record and an output record.
     * as long as the output record < the input record, we continue
     * to loop.
     */
    rv = rwioReadRecord(query_stream, &query_rec);
    if (rv != 0) {
        if (rv == SKSTREAM_ERR_EOF) {
            /* QUERY_FILE is empty */
            have_query = 0;
        } else {
            rwioPrintLastErr(query_stream, rv, &skAppPrintErr);
            exit(EXIT_FAILURE);
        }
    }
    rv = rwioReadRecord(response_stream, &response_rec);
    if (rv != 0) {
        if (rv == SKSTREAM_ERR_EOF) {
            /* RESPONSE_FILE is empty */
            have_response = 0;
        } else {
            rwioPrintLastErr(response_stream, rv, &skAppPrintErr);
            exit(EXIT_FAILURE);
        }
    }

    while (have_query && have_response) {
        rv = compareRecords(&query_rec, &response_rec);
        switch (rv) {
          case -1:
            /* QUERY is too early; read next query */

            /* if we are including unmatched queries, write them out
             * now, after clearing the NHIP field to indicate lack of
             * match */
            if (write_unmatched_query && !printed_query) {
                rwRecSetNhIPv4(&query_rec, 0);
                rv = rwioWriteRecord(matched_stream, &query_rec);
                if (RWIO_ERROR_IS_FATAL(rv)) {
                    rwioPrintLastErr(matched_stream, rv, &skAppPrintErr);
                    exit(EXIT_FAILURE);
                }
            }
            have_query = rwRead(query_stream, &query_rec);
            printed_query = 0;
            break;

          case 0:
            /* RECORDS MATCH.  Write the query record (once), and
             * write all resonse records that match. */
            if (!printed_query) {
                ++match_id;
                rwRecSetNhIPv4(&query_rec, match_id);
                rwioWriteRecord(matched_stream, &query_rec);
                printed_query = 1;
            }
            rwRecSetNhIPv4(&response_rec, 0xFF000000 | match_id);
            rv = rwioWriteRecord(matched_stream, &response_rec);
            if (RWIO_ERROR_IS_FATAL(rv)) {
                rwioPrintLastErr(matched_stream, rv, &skAppPrintErr);
                exit(EXIT_FAILURE);
            }
            have_response = rwRead(response_stream, &response_rec);
            break;

          case 1:
            /* RESPONSE is too early; read next response */

            /* if we are including unmatched responses, write them out
             * now, after settting the NHIP field to indicate lack of
             * match */
            if (write_unmatched_response) {
                rwRecSetNhIPv4(&response_rec, 0xFF000000);
                rv = rwioWriteRecord(matched_stream, &response_rec);
                if (RWIO_ERROR_IS_FATAL(rv)) {
                    rwioPrintLastErr(matched_stream, rv, &skAppPrintErr);
                    exit(EXIT_FAILURE);
                }
            }
            have_response = rwRead(response_stream, &response_rec);
            break;

          default:
            skAppPrintErr("Invalid return from compareRecords()");
            abort();
        }
    }

    /* write the remaining unmatched records */
    if (write_unmatched_query && have_query) {
        if (printed_query) {
            have_query = rwRead(query_stream, &query_rec);
        }
        if (have_query) {
            do {
                rv = rwioWriteRecord(matched_stream, &query_rec);
                if (RWIO_ERROR_IS_FATAL(rv)) {
                    rwioPrintLastErr(matched_stream, rv, &skAppPrintErr);
                    exit(EXIT_FAILURE);
                }
            } while (rwRead(query_stream, &query_rec));
        }
    }

    if (write_unmatched_response && have_response) {
        do {
            rv = rwioWriteRecord(matched_stream, &response_rec);
            if (RWIO_ERROR_IS_FATAL(rv)) {
                rwioPrintLastErr(matched_stream, rv, &skAppPrintErr);
                exit(EXIT_FAILURE);
            }
        } while (rwRead(response_stream, &response_rec));
    }

    if (matched_stream) {
        rwCloseFile(matched_stream);
    }

    /* done */
    appTeardown();

    return 0;
}


/*
** Local Variables:
** mode:c
** indent-tabs-mode:nil
** c-basic-offset:4
** End:
*/
